import visa
import time
import sys
import math
from decimal import Decimal

rotationSecs = 64
dBbeamwidth = -3
readingsPerLine = 15
outputFile = 'pattern.R'

def waitForKey():
    print("Press enter and begin rotation of antenna concurrently ...")
    sys.stdin.readline()
    print("Measuring ...")

def measure(instrument, rotationSecs):
    readings = []
    begin = time.time()
    while (time.time() - begin) < rotationSecs:
        reading = float(instrument.ask("TS MA"))
        angle = ((time.time() - begin) / rotationSecs) * 360
        readings.append([angle, reading])
    return readings

def generateRCode(readings, readingsPerLine):
    code = "library(plotrix)\ndeg <- c("
    degs = [str(Decimal(row[0]).quantize(Decimal('.01'))) for row in readings]
    code += ',\n'.join([','.join(degs[i:i+readingsPerLine]) for i in range(0, len(degs), readingsPerLine)])
    code += ")\nmag <- c("
    mags = [str(Decimal(row[1]).quantize(Decimal('.01'))) for row in readings]
    code += ',\n'.join([','.join(mags[i:i+readingsPerLine]) for i in range(0, len(mags), readingsPerLine)])
    code += ")\npolar.plot(lengths=mag, polar.pos=deg, start=90, clockwise=TRUE, rp.type=\"p\")\n"
    return code

def offsetReadingsBy(readings, by):
    for i in range(0, len(readings)):
        readings[i][1] += by;

def rotateReadingsBy(readings, by):
    for i in range(0, len(readings)):
        readings[i][0] += by;

def findMagnitudePeak(readings):
    peakMag = -sys.float_info.max
    for i in range(0, len(readings)):
        if (readings[i][1] > peakMag):
            peakMag = readings[i][1]
    return peakMag

def findMagnitudePeakAngle(readings):
    peakMag = -sys.float_info.max
    for i in range(0, len(readings)):
        if (readings[i][1] > peakMag):
            peakMag = readings[i][1]
            peakMagStartDeg = readings[i][0]
        if (readings[i][1] == peakMag):
            peakMagEndDeg = readings[i][0]
    return (peakMagStartDeg + peakMagEndDeg) / 2

def findBeamwidth(readings, dB):
    # omnidir pattern may not have a beamwidth!
    # return NaNs in that case
    # This function does not work with multiple disjoint 0dB peaks present
    negdB = float('nan')
    posdB = float('nan')
    peakFound = False
    for i in range(0, len(readings)):
        if (readings[i][1] == 0):
            peakFound = True
        if peakFound:
            if (readings[i][1] <= dB and math.isnan(posdB)):
                posdB = readings[i][0]
        else:
            if (readings[i][1] <= dB):
                negdB = readings[i][0]

    if (math.isnan(negdB) or math.isnan(posdB)):
        return [float('nan'), float('nan'), float('nan')]

    return [posdB - negdB, negdB, posdB]

def findMagnitudeAtAngle(readings, angle):
    lowDeg = float('nan')
    lowMag = float('nan')
    for i in range(0, len(readings)):
        if (readings[i][0] == angle):
            return readings[i][1] # No need to interpolate
        if (readings[i][0] < angle):
            lowDeg = readings[i][0]
            lowMag = readings[i][1]
        if (readings[i][0] > angle):
            # Totally ripped formula from Wikipedia, too tired to verify right now
            # http://en.wikipedia.org/wiki/Interpolation <-- using linear
            return lowMag + (readings[i][1] - lowMag) * ((angle - lowDeg) / (readings[i][0] - lowDeg))
    return float('nan')

sa = visa.instrument('GPIB0::18')
sa.write("O3")
waitForKey()

readings = measure(sa, rotationSecs)

print("Detected measurement resolution of {0} degrees.".format(float(360) / len(readings)))

print("Finding peak magnitude and angle ...")
peakMag = findMagnitudePeak(readings)
#peakDeg = findMagnitudePeakAngle(readings) # Bad method, center graph on center of beamwidth instead

print("Normalizing magnitude by adding {0}dB ...".format(-peakMag))
offsetReadingsBy(readings, -peakMag)

beamwidth = findBeamwidth(readings, dBbeamwidth)
print("{0}dB beamwidth is {1} degrees.".format(dBbeamwidth, beamwidth[0]))

beamwidthCenterDeg = (beamwidth[1] + beamwidth[2]) / 2
print("Centering on beamwidth by rotating readings {0} degrees ...".format(-beamwidthCenterDeg))
rotateReadingsBy(readings, -beamwidthCenterDeg)

print("-120 degree gain = {0}dB, +120 degree gain = {1}dB.".format(findMagnitudeAtAngle(readings, -120),
                                                                   findMagnitudeAtAngle(readings, 120)))

print("Writing results to {0}".format(outputFile))
with open(outputFile, 'w') as f:
   f.write(generateRCode(readings, readingsPerLine))

print("Done!")
